class CoreHeap():
    def __init__(self):
        self.data = {}

    def __len__(self):
        return len(list(self.data.keys()))

    def  __contains__(self, node):
        return node in self.data

    def add(self, node):
        if not node in self:
            self.data[node] = node
        return self.data[node]

    def clear(self):
        self.data = {}

    @property
    def asDict(self):
        return self.data

    @property
    def dimensions(self):
        lnkNum = 0
        for k,v in self.data.items():
            lnkNum += len(v['children']) 
        return len(self), lnkNum
    
    def __iter__(self):
        for k, v in self.data.items():
            yield v

    def items(self):
        for k, v in self.data.items():
            yield (k,v)       

class FHeap(CoreHeap):
    def __init__(self):
        super().__init__()
    def add(self, cNode, pNode):
        if not cNode.ID in self.data:
            self.data[cNode.ID] = {
                'name' : cNode.name,
                'eTag' : cNode.eTag,
                'children' : [ _.ID for _ in cNode.children ],
                'is_a' : [],
                'isDAGelem' : cNode.isDAGelem
            }
        if not pNode is None:
            self.data[cNode.ID]['is_a'].append(pNode.ID) # Do we check unicity?

class VHeap(CoreHeap):
    def __init__(self):
        super().__init__()
        self.totalElements = set()
    def add(self, cNode, pNode):
        if not cNode.ID in self.data:
            _  = cNode.getMembers(nr=True)
            self.data[cNode.ID] = {
                'name' : cNode.name,
                'elements' : _,
            }
            self.totalElements = self.totalElements | set(_)
        if not pNode is None:
            self.data[cNode.ID]['is_a'].append(pNode.ID) # Do we check unicity?
    
    @property
    def asDict(self):
        registry = list(self.totalElements)

        _ = { "registry"  : registry,
              "terms"     : {}
            }
        
        for goID, goTerm in self.items():
            _["terms"][goID] = {
                "name" : goTerm['name'],
                "elements" : [ registry.index(e) for e\
                                in  goTerm["elements"] ],
                "freq"   : float( len(goTerm["elements"])/ len(registry) )  
            }
        return _



from flask import Flask, request
import data_objects

def hello():
    data = request.get_json()
    schema = data_objects.CulledGoParametersSchema()
    parameters = schema.load(request)
    print(data)
    print(parameters)
    return 'hello'

app = Flask(__name__)
app.add_url_rule("/", 'hello', hello, methods=['POST'])

app.run(debug=True, port=3333)


from marshmallow import Schema, fields, pre_load
from pyproteinsExt.uniprot import isValidID
from ..tree import enumNS
from .. import Univgo as createUnivgo

DEFAULT_MIN_COUNT = 0
DEFAULT_MAX_COUNT = 50
DEFAULT_MAX_FREQ = 1.0

def validateCount(n):
    if n < 0:
        raise ValidationError("Count must be greater than 0.")

def validateFreq(n):
    if n < 0:
        raise ValidationError("maxFreq must be greater than 0.")
    if n > 1.0:
        raise ValidationError("maxFreq must be lower or equal to 1.")
    
class CulledGoParameters:
    def __init__(self, minCount, maxCount, maxFreq):
        self.minCount = int(minCount)
        self.maxCount = int(maxCount)
        self.maxFreq  = float(maxFreq)
    def __repr__(self):
        return f'<CulledGoParameters(minCount={self.minCount}, maxCount={self.maxCount}, maxFreq={self.maxFreq})>' 

class CulledGoParametersSchema(Schema):
    global DEFAULT_MIN_COUNT, DEFAULT_MAX_COUNT, DEFAULT_MAX_FREQ

    @pre_load(pass_many=True)
    def unwrap_request(self, request, **kwargs):
        return request.get_json()

    minCount = fields.Int(validate=validateCount,  missing=DEFAULT_MIN_COUNT)
    maxCount = fields.Int(validate=validateCount,  missing=DEFAULT_MAX_COUNT)
    maxFreq  = fields.Float(validate=validateFreq, missing=DEFAULT_MAX_FREQ)

def validateUnivgoSerial(fn):
    
    def loadUnivGO_dec(data):
        for k in ["single_tree", "omega_uniprotID", "ns"]:
            if k not in data:
                raise KeyError(f"univoGo serial key {k} missing")
    
        if not type(data["omega_uniprotID"]) == list:
            raise KeyError(f"univGo serial key omega_uniprotID is not a list\n=>{type(data['omega_uniprotID'])}")
        for uID in data["omega_uniprotID"]:
            if not isValidID(uID):
                raise ValueError(f"univGo serial uniprotID not valid {uID}")
        if data["ns"] not in enumNS:
            raise ValueError(f"univGo serial NS not valid {data['ns']}")
        return fn(data)
        
    
    return loadUnivGO_dec

@validateUnivgoSerial
def loadUnivGO(d):
    univGo = createUnivgo(serial=d)
    return univGo



"""
Basic statistical functions to compute pathway enrichment

"""

from scipy.stats import hypergeom
from scipy.stats import fisher_exact
import numpy as np
from scipy.cluster.hierarchy import dendrogram, ward, fcluster

def computeSelfORA(node, proteinList):
    """computeSelfORA
    Compute the probability of observing at least the k members of the supplied GOnodes/pathway
    present in the supplied protein list
    
    Parameters
    ----------
    node : element of the ontology tree.
    proteinList: list or uniprot identifiers
    
    Remarks
    ----------
    Background proteins are all uniprot ID found in the supplied annotation subtree
    """

    ORA = []
    
    # universe is all uniprotID found in the annotation tree
    universe = set(node.getMembers()) 
    N = len(universe)
    # nSet is the observed set   
    nSet = set(proteinList)
    n = len(nSet)
    for cPath in node.walk():       
        Kstates = set(cPath.getMembers())
        K = len( Kstates )
        print(f"{cPath.name} has {K} members")
        
        k_obs = Kstates & nSet
        k = len(k_obs)
        
        p = righEnd_pValue(N, n, K, k)
        
        ORA.append( (p, cPath) )
        print(f"{cPath.name} [{K} -> {k}/{n}] = {p}")
    return ORA
    
def righEnd_pValue(N, n, K, k):
    
#print(f"N={N}, n={n}, K={K}, k={k}")

#The hypergeometric distribution models drawing objects from a bin. 
#N is the total number of objects, K is total number of Type I objects. 
#The random variate represents the number of Type I objects in N 
#drawn without replacement from the total population.

# Right-end tail of the CDF is P(X>=k)
    p_x = hypergeom(N, K, n).cdf([k - 1])
    return 1.00 - p_x[0]

def computeORA(node, proteinList, nodeBKG, verbose=False):
    Fisher, CDF = computeORA_BKG(node, proteinList, nodeBKG, verbose=verbose)
    return Fisher

def computeORA_BKG(node, proteinList, nodeBKG, verbose=False): # IDEM, mais avec un autre arbre de reference
    
    if verbose:
        print("Computing Over Representation w/ a background tree")

    ORA_Fisher = []
    ORA_CDF = []

    universe = set(nodeBKG.getMembers())
    o = len(universe)

    # nSet is the observed set   
    nSet = set(proteinList)
    n = len(nSet)
    
    pathwayPotential = 0
    pathwayReal = 0
    
    import time

    start = time.time()
   # for cPath in node.walk(mustContain=proteinList):
    for cPath in node.walk():
        pathwayPotential += 1
        #verbose = cPath.name == 'enzyme binding'
        if verbose:
            print(cPath.name)

        # Table de contingence
        #
        #        | Pa  | non_PA |
        # -----------------------
        #    SA  |     |        |
        #  nonSA |     |        |

        
        # l'intersection entre les protéines porteuse de l'annotation courante et 
        # la liste des protéines sur-abondante
        # => nombre de succès observés, "k"
        Kstates = set(cPath.getMembers())
        k_obs = Kstates & nSet
        if not k_obs:
            if verbose:
                print("k_obs == 0")
            continue
        k = len(k_obs)
        pathwayReal += 1
        
        # Pour estimer le nombre de protéines non surAbondantes appartenant au pathway ou non
        # Nous utilisons la proporition de protéines du pathway ou non dans le protéome entier
        bkgPath = nodeBKG.getByName(cPath.name)
        bkgPathFreq = len( set(bkgPath.getMembers()) ) / len(universe)  # Fraction du protéomes appartenant à ce Pathway
        nSA_Pa = int ( (o - k) * bkgPathFreq )
        nSA_nPa = int( (o - k) - nSA_Pa )
        
        TC = [
            [ k ,  len(proteinList) - k],
            [ nSA_Pa ,  nSA_nPa]
        ]
        
        oddsratio, pValue = fisher_exact(TC, alternative="greater")
        p = righEnd_pValue(o, n, len( set(bkgPath.getMembers()) ), k)

        if verbose:
            print(f"{cPath.name} {TC} p={pValue} // pL={p}")

        
        ORA_Fisher.append( (pValue, cPath) ) 
        ORA_CDF.append( ( p, cPath) ) 
        
        cPath.set(Fisher=pValue, Hpg=p)

    end = time.time()    
    print(f"Evaluated {pathwayReal} / {pathwayPotential} Pathways, based on {n} proteins in {end - start} sc")
    return ORA_Fisher, ORA_CDF

def applyOraToVector(vectorizedProteomeTree, experimentalProteinID, deltaProteinID, threshold=0.005):
    """ compute Fischer exact test on a list of datastructure corresponding 
        to a vectorized full proteome GO tree.

        Parameters
        ----------
        vectorizedProteomeTree : Univgo.vectorize() return datastructure
        experimentalProteinID  : List of uniprot identifiers of observed proteins
        deltaProteinID         : List of uniprot identifiers of over/under expressed proteins
    """

    def ora(universe, pathway, observedProteinList, deltaProteinList):
        o       = len(universe)
        nSet    = set( observedProteinList )
        n       = len(nSet)
        Kstates = set( pathway["elements"] )
        k_obs   = Kstates & set(deltaProteinList)
        k = len(k_obs)
        
        nSA_Pa  = int ( (o - k) *  pathway["freq"] )
        nSA_nPa = int( (o - k) - nSA_Pa )
            
        TC = [
            [ k ,  n - k],
            [ nSA_Pa ,  nSA_nPa]
        ]

        oddsratio, pValue = fisher_exact(TC, alternative="greater")
        return {
            "name"       : pathway["name"],
            "pvalue"     : pValue,
            "K_states"   : pathway["elements"],
            "k_success"  : list(k_obs),
            "table"      : TC,
            "bkgFreq"    : pathway["freq"]
        }

    d = vectorizedProteomeTree
    registry = d["registry"]
    # delta _include_in experimental _include_in wholeProteome
    
    # B/C vectors are not guaranted to be universal, protein may not be found in registry
    # assert( not set(experimentalProteinID) - set(registry)              )
    assert( not set(deltaProteinID)        - set(experimentalProteinID) )
    annotatedExperimentalProteinID = list( set(experimentalProteinID) & set(registry) )
    annotatedDeltaProteinID        = list( set(deltaProteinID)        & set(registry) )

    #print("REGISTRY::: " ,d["registry"])
    # Convert two uniprotID list to integers
    expUniprotIndex   = [ d["registry"].index(_) for _ in annotatedExperimentalProteinID ]
    deltaUniprotIndex = [ d["registry"].index(_) for _ in annotatedDeltaProteinID        ]
    
    res = { goID: ora(d['registry'], ptw, expUniprotIndex, deltaUniprotIndex) for goID, ptw in d['terms'].items() }

    return { k:v for k,v in res.items() if v["pvalue"] < threshold }




def kappaClustering(registry, applyOraToVectorResults, fuseThresh=0.2):
    """ Cluster Pathways evaluated by applyOraToVector scoring using Cohen's kappa coefficient
    """
    omega = set( [ k for k in range(len(registry)) ] )
    pathwayID = list( applyOraToVectorResults.keys() )

    for x in range(0, len(pathwayID)):
        _ = applyOraToVectorResults[ pathwayID[x] ]
    #    print(x," | ", _["name"], _["K_states"])



    #pathwayID = pathwayID + pathwayID
    pdist = []  
    for i in range(0, len(pathwayID) - 1):
        iID, iTerm   = (pathwayID[i],  applyOraToVectorResults[ pathwayID[i] ])
        iName        = iTerm["name"]

        in_iTerm     = set(iTerm["K_states"])
        not_in_iTerm = omega - in_iTerm
        for j in range(i + 1, len(pathwayID)):
            jID, jTerm = (pathwayID[j],  applyOraToVectorResults[ pathwayID[j] ])
            jName        = jTerm["name"]

            in_jTerm     = set(jTerm["K_states"])
            not_in_jTerm = omega - in_jTerm
    #        print(f"\n\nGO Term [{i}]{iName} [{j}]{jName}")
    #        print(in_iTerm, in_jTerm)#, not_in_iTerm)
            k = kappa( in_iTerm     & in_jTerm    ,\
                       in_iTerm     & not_in_jTerm,\
                       not_in_iTerm & in_jTerm    ,\
                       not_in_iTerm & not_in_jTerm ,\
                )
            
    #        print(f"pdist = {k}")
            pdist.append( 1 - k)

    Z = ward(np.array(pdist))
    #print(Z)
    _Z = fcluster(Z, t=fuseThresh, criterion='distance')
    #print(_Z)

    V = flattenToD3hierarchy(_Z.tolist(), registry, applyOraToVectorResults, pathwayID)
    print(V)

    return V

def kappa(a, b, c, d):
    """           GO term 2
               | yes |  no |        
    -------------------------------   
    GO   | yes |  a  |  b  | 
   term1 |  no |  c  |  d  |


   kapa(GO_1, GO_2) = 1 - (1 - po) / (1 - pe)

   po = (a + d) / (a + b + c + d) 
   marginal_a = ( (a + b) * ( a + c )) / (a + b + c + d)
   marginal_b = ( (c + d) * ( b + d )) / (a + b + c + d)
   pe = (marginal_a + marginal_b) / (a + b + c + d)

"""
    a = float(len(a))
    b = float(len(b))
    c = float(len(c))
    d = float(len(d))

    po = (a + d) / (a + b + c + d) 
    marginal_a = ( (a + b) * ( a + c )) / (a + b + c + d)
    marginal_b = ( (c + d) * ( b + d )) / (a + b + c + d)
    pe = (marginal_a + marginal_b) / (a + b + c + d)

    #print (f" {a} | {b}\n {c} | {d}")
    return 1 - (1 - po) / (1 - pe)


def flattenToD3hierarchy(_Z, registry, applyOraToVectorResults, pathwayID):
    clusterElement = { }
    for pathwayIndex, clusterNum in enumerate(_Z):
        currPathwayID    = pathwayID[pathwayIndex]
        currPathway      = applyOraToVectorResults[currPathwayID]
        if not clusterNum in clusterElement:
            clusterElement[clusterNum] = {
                "name": None,
                "children":[],
                "best" : 1.1
            }
        currCluster = clusterElement[clusterNum]
        proxy = { k : v for k,v in currPathway.items() if not (k == "K_states" or  k == "k_success") }
        currCluster["children"].append(proxy)
        proxy["uniprotID"] = [ registry[int(_uid)] for _uid in currPathway["k_success" ] ]
        proxy["maxMemberCount"] = len(currPathway["K_states"])
        if currPathway["pvalue"] < currCluster["best"]:
            currCluster["best"] = currPathway["pvalue"]
            currCluster["name"] = currPathway["name"] + "_group"

    return {
        "name" : "root",
        "children" : [ clust for _, clust in clusterElement.items() ]
    }



import requests
from pyproteinsExt import uniprot as pExt
from . import Univgo as createGOTreeUniverse

PROXIES = {
	'http': '',
	'https': '',
}

def setProxy(http, https):
	global PROXIES
	PROXIES["http"] = http
	PROXIES["https"] = https
  
def check_proteins_subset(major_list:list[str], sublist:list[str]) -> bool:
	return set(sublist).issubset(major_list)

def unigo_tree_from_api(api_host:str, api_port:int, taxid:int) -> str:
	'''Interrogate GO store API and return requests response'''
	go_url = f"http://{api_host}:{api_port}/unigo/{taxid}"
	print(f"Interrogate {go_url} for go tree")
	return requests.get(go_url, proxies = PROXIES)

def unigo_vector_from_api(api_host:str, api_port:int, taxid:int) -> str:
	'''Interrogate GO store API and return requests response'''
	go_url = f"http://{api_host}:{api_port}/vector/{taxid}"
	print(f"Interrogate {go_url} for go vector")
	return requests.get(go_url, proxies = PROXIES)

def unigo_culled_from_api(api_host:str, api_port:int, taxid:int, goParameters:{}):
	'''Interrogate GO store API and return requests response'''
	go_url = f"http://{api_host}:{api_port}/vector/{taxid}"
	print(f"Interrogate {go_url} for go culled vector {goParameters}")
	return requests.post(go_url, proxies = PROXIES, json=goParameters)


def loadUniprotCollection(proteomeXML, strict=True):
	"""Parse provided Uniprot XML and return following 2-uple
		: (<first_taxid_in_collection>, <Uniprot_Collection>)
	"""
	print(f"Loading a protein collection from {proteomeXML}")

	uColl = pExt.EntrySet(collectionXML=proteomeXML)
	_ = uColl.taxids
	if len(_) != 1 and strict:
		raise ValueError(f"Taxids count is not equal to 1 ({len(_)}) in uniprot collection : {_}")
	if len(_) != 1 and not strict:
		print(f"Warning: Taxids count is not equal to 1 ({len(_)}) in uniprot collection : {_}")
	
	return _[0], uColl

    
def generateDummySet(uColl, nTotal, deltaFrac = 0.1):
	print(f"Setting up a dummy experimental collection of {nDummy} elements")
	expUniprotID =[]
	for uObj in uColl:
		if len(expUniprotID) == nTotal:
			break
		if uObj.isGOannot and uObj.taxid == uTaxid:
			expUniprotID.append(uObj.id)
	
	nDelta = int(nDummy*deltaFrac)
	print(f"Considering {nDelta} proteins among {nTotal} experimental as of significantly modified quantities")
	return expUniprotID[:nDelta]

def loadUniprotIDsFromCliFiles(expressedProtIDFile, deltaProtIDFile):
	expUniprotID   = []
	deltaUniprotID = []
	with open(expressedProtIDFile) as f:
		for l in f:
			expUniprotID.append(l.rstrip())
	with open(deltaProtIDFile) as f:
		for l in f:
			deltaUniprotID.append(l.rstrip())

	if not check_proteins_subset(expUniprotID, deltaUniprotID):
		raise Exception("Differentially expressed proteins are not completely included in total proteins")

	return expUniprotID, deltaUniprotID


def loadUniversalTreesFromXML(proteomeXMLs, owlFile):
	print(f"Loading following XML proteome(s) {proteomeXMLs}\n This may take a while...")

	uTaxids = []
	uTrees  = []
	for proteomeXML in proteomeXMLs:
		uTaxid, uColl = loadUniprotCollection(proteomeXML)
		uTaxids.append(uTaxid)
		tree_universe = createGOTreeUniverse( 
											owlFile     = owlFile,
											ns          = "biological process", 
											fetchLatest = False,
											uniColl     = uColl)
		uTrees.append(tree_universe)
	
	return zip(uTaxids, uTrees)


def parseGuessTreeIdentifiers(identifersList):
	""" Returns a list of tree key identifiers guessed from input list
		Input list elements can be actual tree keys or proteomeXML files
	"""
	guessedTreeID = []
	for ressourceMaybeXML in identifersList:
		try:
			_, uColl = loadUniprotCollection(ressourceMaybeXML)
			guessedTreeID += uColl.taxids
		except FileNotFoundError:
			#print(f"{ressourceMaybeXML} does not look like a file")
			guessedTreeID.append(ressourceMaybeXML)
		except IsADirectoryError:
			print(f"Warning {ressourceMaybeXML} looks like a directory")
			guessedTreeID.append(ressourceMaybeXML)
		except ValueError:
			raise IOError(f"{ressourceMaybeXML} looks like an invalid proteome XML file")
		#except Exception as e:
		#	print(type(e).__name__)
	return list(set(guessedTreeID))


from flask import request, abort
from .. import utils
def checkPwasInput(enforcedCulling=True):
    data = request.get_json()

    if not data:
        print(f"ERROR in request : empty posted data. Abort 400")
        abort(400)
    
    for key in ["all_accessions", "taxid", "significative_accessions", "method"]:
        if not key in data:
            print(f"ERROR in request : {key} not in posted data. Abort 400")
            abort(400)

    if not data["method"] in ["fisher"]:
        print(f"ERROR : this statistical method ({data['method']}) is not handled. Availables : fisher")
        abort(400)        

    if not utils.check_proteins_subset(data["all_accessions"], data["significative_accessions"]):
        print(f"ERROR : significative accessions are not all included in all accessions")
        abort(400)
        
    return data


import requests

HOSTNAME=None
PORT=None

def handshake(hostname, port):
    try:
        url = f"http://{hostname}:{port}/ping"
        req = requests.get(url)
    except ConnectionError as e:
        raise ConnectionError(f"Unable to handshake at {url}\n->{e}")

    if not req.status_code == requests.codes.ok:
        raise ConnectionError(f"Error {req.status_code} while handshaking at {url}")
        if not req.text == "pong":
            raise ConnectionError(f"Improper handshake ({req.text}) at {url}")
    
    global HOSTNAME, PORT
    
    HOSTNAME = hostname 
    PORT     = port
    return True

def addTreeByTaxid(treeTaxidIter):
    for taxid, tree in treeTaxidIter:
        d = tree.serialize()
        url = f"http://{HOSTNAME}:{PORT}/add/unigo/{taxid}"
        req = requests.post(url, json=d)
        if req.status_code == requests.codes.ok:
            print(f"Successfull tree adding at {url}")
        else:
            print(f"Error {req.status_code} while inserting at {url}")

def delTreeByTaxid(taxids):
    print(f"Want to del by taxids {taxids}")
    for taxid in taxids:
        url = f"http://{HOSTNAME}:{PORT}/del/unigo/{taxid}"
        req = requests.delete(url)
        if req.status_code == requests.codes.ok:
            print(f"Successfull tree deleting at {url}")
        else:
            print(f"Error {req.status_code} while deleting at {url}")

def buildVectors():
    """ Trigger Vector building
    """
    url = f"http://{HOSTNAME}:{PORT}/list/trees"
    req = requests.delete(url)


import re, json, copy

from .heap import CoreHeap as kNodes 

class Node():

    def __init__(self, ID, name, oNode=None):
        self.ID = ID 
        self.name = name
        self.eTag =  [] # List of elements actually carrying the annotation ("tagged by the nodeName/annotation")
        self.leafCount =  0
        self.children =  []
        self.features = {}
        self.oNode = oNode
        self.isDAGelem = False
        self.is_a = [] # Used to deserialize from api
        #self.heap = None

    def __deepcopy__(self, memo):
        # Deepcopy only the id attribute, then construct the new instance and map
        # the id() of the existing copy to the new instance in the memo dictionary
        memo[id(self)] = newself = self.__class__(copy.deepcopy(self.ID, memo), copy.deepcopy(self.name, memo), copy.deepcopy(self.oNode, memo))
        # Now that memo is populated with a hashable instance, copy the other attributes:
        newself.eTag = copy.deepcopy(self.eTag, memo)
        # Safe to deepcopy now, because backreferences to self will
        # be remapped to newself automatically
        newself.children = copy.deepcopy(self.children, memo)
        newself.features = copy.deepcopy(self.features, memo)
        newself.isDAGelem = copy.deepcopy(self.isDAGelem, memo)
        newself.oNode = copy.deepcopy(self.oNode, memo)
        #newself.heap = copy.deepcopy(self.heap, memo)

        return newself


    #def serial(self):


    def __hash__(self):
        return hash(self.ID)
    
    def __eq__(self, other):
        return hash(self) == hash(other)

# We may have to memo these 3   
    def as_DAG(self):
        return {
            'id' : self.ID,
            'children' : [ c.as_DAG() for c in self.children ]
        }

    def _as_DAG_strat(self, flat):
        if len(self.children) == 0 :
            return

        for c in self.children:
            if not c.ID in flat:
                flat[c.ID] = { "id" : c.ID, "parentIds" : set() }
            flat[c.ID]["parentIds"].add(self.ID)
        
        for c in self.children:
            c._as_DAG_strat(flat)
    
    def as_DAG_strat(self):
        #stHeap = kNodes()
        flat = { self.ID : { "id": self.ID, "parentIds" : [] } }
        #stHeap.add(self)

        self._as_DAG_strat(flat)
        
        return [ { "id" : v["id"], "parentIds" : list( v["parentIds"] )}  for k,v in flat.items() ]

    @property
    def pvalue(self):
        if 'Fisher' in self.features:
            return self.features['Fisher']
        return None
        

    def set(self, **kwargs):
        for k,v in kwargs.items():
            self.features[k] = v

    def __getattr__(self, key):
      #  print(key)
    #   if key not in self.features:
        if key == "ID":
            print(dir(self))
            return self.__getattribute__(key)

        if key == "features":
            raise AttributeError(key)
        try:
            return self.features[key]
        except KeyError:
            raise AttributeError(key)

        #return self.features[key]
    
    def __dir__(self):
        return super().__dir__() + [str(k) for k in self.features.keys()]
    
    def hasChild(self, ID):
    # print(node)
        for child in self.children:
            if child.ID == ID:
                return child
        return None
    
   
    def traverse(self):
        yield self
        for c in self.children:
            yield from c.traverse()
     
    # Memoized version of traverse
    def walk(self, **kwargs):
        wHeap = kNodes()
        return self._walk(wHeap, **kwargs)

    def _walk(self, wHeap, **kwargs):
        if self.isDAGelem and self in wHeap:            
            return        
        wHeap.add(self) 
        if 'mustContain' in kwargs:
            if not set(kwargs["mustContain"]) & set(self.getMembers()) :
                #print(f"-->Early exit at empty node {self.name}")
                return

        yield self
        for c in self.children:
            yield from c._walk(wHeap, **kwargs)

    def _as_newick(self):
        #_self = node['name'].replace(" ", "_")
        _self = self.name.replace(',',' ').replace('(', '[').replace(')', ']').replace('\'', '_').replace(':', '_')
        
        if len(self.children) == 0:
            return _self

        return '(' + ','.join([ c._as_newick() for c in self.children ]) + ')' + _self
            
    def getByName(self, name):       
        regExp = name.replace(' ', '.').replace('[', '.').replace(']', '.').replace('_', '.').replace(')', '\)').replace('(', '\(')
        regExp = regExp.replace('+', '.')
        if re.search("^" + regExp + "$", self.name):
            return self
        for c in self.children:
            v = c.getByName(name)
            if v:
                return v
        return None

    def getByID(self, ID):       
        if self.ID == ID:
            return self
        for c in self.children:
            v = c.getByID(ID)
            if v:
                return v
        return None


# Memoization implies the leaves of the subtree are already registred
# API to perform search at current node, we clear the heap

    def getMembers(self, nr=False):
        getMemberHeap = kNodes()
        getMemberHeap.add(self)

        buff = copy.copy(self.eTag)
        for n in self.children:
            buff += n._getMembers(getMemberHeap)
        
        return buff if not nr else list(set(buff))

    def _getMembers(self, _heap):
        if self.isDAGelem and self in _heap:
            return []        
        _heap.add(self)
        buff = copy.copy(self.eTag)
        for n in self.children:
            buff += n._getMembers(_heap)
        
        return buff


    #def _mayCollapseNode(self, fnPredicate, _ctHeap):
    #""" Apply a predicate to collapse a node
    #    Collapsing a node makes 
    #"""


    def _collapseNode(self, _ctHeap):
        #if self.isDAGelem and self in _ctHeap:
        #    print(f"{self.name} already visited") ## This is buggy
        #    return self
        
        #print(f"Looking in {self.name}")
        #for n in self.children:
        #    print(f"{self.name} -> {n.name}")
        _ctHeap.add(self)
        #print(f"collapsing {self.name}")
        if len(self.children) == 0: # Leave or a node carrying actual protein, return it
            return self
        
        if len(self.children) == 1 and len(self.eTag) == 0 :
         #   print(f"Skipping {self.name}")
            return self.children[0]._collapseNode(_ctHeap)

        #print(f"Keeping {self.name} children={len(self.children)} eTag{len(self.eTag)}")
        self.children = [ n._collapseNode(_ctHeap) for n in self.children ]
        
        return self

    def __repr__(self):
        d = { k : v for k,v in self.__dict__.items()  if not k == "children" }
        d["children"] = [ c.name for c in self.children ]
        return str(d)

    def __str__(self):
        return self.__repr__()

    def _mayDrop(self, predicate, _noDropHeap):
        #print(f"Testing {self.name}")
        if not predicate(self):
        #   print(f"{self.name} is droped !")
            return None

        if self in _noDropHeap:
            return self
        _noDropHeap.add(self)

        self.children = [ c for c in self.children if c.mayDrop(predicate, _noDropHeap) ]

        return self

    def mayDrop(self, predicate, pNode, _noDropHeap):
        # Self failed add nothing to stack
        if not predicate(self):        
            return
        # add self to stack
        _noDropHeap.add(self, pNode)

        # pass on to children
        for c in self.children:
            c.mayDrop(predicate, pNode, _noDropHeap)

        return

    def _leafCountUpdate(self, _lcHeap):
      #  print(f"lcu {self.name} {len(_lcHeap)}")
        if self.isDAGelem and self in _lcHeap:
            #print(f"{self.name} already updated")
            return self
        _lcHeap.add(self)

        self.leafCount = len(self.getMembers())
        #print(f"{self.name} leafCount is {self.leafCount}")

        for c in self.children:
            c._leafCountUpdate(_lcHeap)



from flask import Flask, abort, jsonify, request
import requests, json
from marshmallow import EXCLUDE
from .. import vloads as createGOTreeTestUniverseFromAPI
from .. import uloads as createGOTreeTestFromAPI
from .. import utils
from .io import checkPwasInput
from ..stat_utils import applyOraToVector, kappaClustering
from .data_objects import CulledGoParametersSchema as goParameterValidator

GOPORT=1234
GOHOST="127.0.0.1"
def listen(goApiHost:str, goApiPort:int, vectorized:bool):
    global GOPORT, GOHOST
    GOPORT = goApiPort
    GOHOST = goApiHost

    app = Flask(__name__)
    app.add_url_rule("/", 'hello', hello)
    if vectorized:
        print(f"PWAS API vector listening")
        app.add_url_rule("/compute", "computeOverVector", computeOverVector, methods=["POST"])
    else:
        print(f"PWAS API tree listening")
        app.add_url_rule("/compute", "computeOverTree", computeOverTree, methods=["POST"])
    return app

def hello():
    return "Hello pwas"

def computeOverVector():
    forceUniversal = False
    data = checkPwasInput()
   
    print(f'I get data with {len(data["all_accessions"])} proteins accessions including {len(data["significative_accessions"])} significatives')
    
    if forceUniversal:
        go_resp = utils.unigo_vector_from_api(GOHOST, GOPORT, data["taxid"])
    else:
         # Culling vector parameters
        _goParameterValidator = goParameterValidator()
        goParameter = _goParameterValidator.load(request,  unknown=EXCLUDE)
        go_resp = utils.unigo_culled_from_api(GOHOST, GOPORT, data["taxid"], goParameter)

    if go_resp.status_code != 200:
        print(f"ERROR request returned {go_resp.status_code}")
        abort(go_resp.status_code)
    
    vectorizedProteomeTree = json.loads(go_resp.text)
    #print(vectorizedProteomeTree)

    res = applyOraToVector(vectorizedProteomeTree, data["all_accessions"], data["significative_accessions"], 0.05)
    print(f"clustering {len(res.keys())} GO terms")
    #print("NO Clustering")
    #return jsonify(res)
    Z = kappaClustering(vectorizedProteomeTree["registry"], res)

    return jsonify(Z)

def computeOverTree():
    data = checkPwasInput() 
    print(f'I get data with {len(data["all_accessions"])} proteins accessions including {len(data["significative_accessions"])} significatives')

    go_resp = utils.unigo_tree_from_api(GOHOST, OPORT, data["taxid"])

    if go_resp.status_code != 200:
        print(f"ERROR request returned {go_resp.status_code}")
        abort(go_resp.status_code)
    
    print("Create Unigo tree")
    unigoTreeFromAPI = createGOTreeTestFromAPI(go_resp.text, data["all_accessions"])
    x,y = unigoTreeFromAPI.dimensions
    print("Unigo Object successfully buildt w/ following dimensions:")
    print(f"\txpTree => nodes:{x[0]} children_links:{x[1]}, total_protein_occurences:{x[2]}, protein_set:{x[3]}")  
    print(f"\t universeTree => nodes:{y[0]} children_links:{y[1]}, total_protein_occurences:{y[2]}, protein_set:{y[3]}")  

    #Compute fisher stats
    if data["method"] == "fisher":
        print("Computing ORA with fisher")
        rankingsORA = unigoTreeFromAPI.computeORA(data["significative_accessions"], verbose = False)
        
        return rankingsORA.json

    return {"not computed": "unavailable stat method"}




"""Go ontology tree manipulation tool and microservice

Usage:
    unigo store server redis start [clear] [<owlFile> <xmlProteomeFile>... ] [--rh=<redis_host> --rp=<redis_port> --go=<store_port>]
    unigo store server local start <xmlProteomeFile>... [--gp=<store_port>]
    unigo store client add  <owlFile> <xmlProteomeFile>... [--gp=<store_port> --gh=<store_host>]
    unigo store client del (<xmlProteomeFile>...|<taxids>...) [--gp=<store_port> --gh=<store_host>]  
    unigo pwas server (vector|tree) [--pwp=<pwas_port> --gp=<store_port> --gh=<store_host>]
    unigo pwas cli (vector|tree) <taxid> <expressed_protein_file> <delta_protein_file> [--gp=<store_port> --gh=<store_host> --method=<statMethod>]
    unigo pwas test local  (tree|fisher|convert) [(<xmlProteomeFile> <owlFile>)] [--size=<n_exp_proteins> --delta=<n_modified_proteins> --head=<n_best_pvalue>]
    unigo pwas test client [--port=<portNumber>] [--method=<statMethod>] [--prot=<xmlFile>] 

Options:
  -h --help     Show this screen.
  <owlFile>  ontology file location in owl format.
  <xmlProteomeFile>  uniprot file location in xml format.
  --size=<number>  number of uniprotID to build the experimental protein group from the uniprot elements [default: 50]
  --delta=<number>  fraction of the experimental protein group to build the group of over/undee-represented protein [default: 0.1]
  --silent  stop ORA scoring dump
  --head=<n_best_pvalue>  display n best GO pathway [default: 5]
  --gp=<store_port>  port for GO API [default: 1234]
  --gh=<host>  host name for GO API [default: localhost]
  --pwp=<pwas_port>  port for pwas API [default: 5000].
  --method=<statMethod>  statistical method to compute pathway p-value [default: fisher]
  <expressed_protein_file>  txt file with all proteomics experience proteins accession (one per line)
  <delta_protein_file>  txt file with all differentially expressed proteins accession (one per line)
  --rp=<redis_port>  redis DB TCP port [default: 6379]
  --rh=<redis_host>  redis DB http adress [default: localhost]
"""
import os

from docopt import docopt


#from . import vloads as createGOTreeUniverseFromAPI

from .api.store import bootstrap as goStoreStart

from .api.store.client import addTreeByTaxid as goStoreAdd
from .api.store.client import delTreeByTaxid as goStoreDel
from .api.store.client import handshake


from .utils import loadUniversalTreesFromXML, parseGuessTreeIdentifiers

from .api.pwas import listen as pwas_listen

if __name__ == '__main__':
    arguments = docopt(__doc__)
    print(arguments)

    nDummy      = int(arguments['--size'])
    nTop        = int(arguments['--head'])
    goApiPort   = arguments['--gp'] 
    goApiHost   = arguments['--gh'] 
    pwasApiPort = arguments['--pwp']
    method      = arguments['--method']

    if arguments['store']:
        #Load or create proteins sets
        # With default inputs for proteome and annotations
        #              settings 
        # TO RUN & VALIDATE
        if arguments['test']:
            target = list(set(['fisher', 'convert', 'tree']) & set(arguments.keys()))[0]
            test.run(nDummy, nTop, target=target)
            exit(0)
    
        if arguments['server']: # Provider-Ressource bootstrap API
            if not arguments["<xmlProteomeFile>"]:
                print(f"Resuming goStore service")
                taxidTreeIter = None 
            else:
                print(f"Starting goStore service with proteome{arguments['<xmlProteomeFile>']} as a universe GO tree")
                taxidTreeIter = loadUniversalTreesFromXML(\
                    arguments["<xmlProteomeFile>"],\
                    arguments["<owlFile>"])
            
            app = goStoreStart(newElem=taxidTreeIter,\
                clear = True if arguments['clear'] else False,\
                cacheType='local' if not arguments['redis'] else 'redis',\
                rp=arguments['--rp'], rh=arguments['--rh'],\
                _main_ = __name__ == '__main__' )  
            app.run(debug=False, port=goApiPort)

        elif arguments['client']: # Client-Ressource mutation API
            handshake(goApiHost, goApiPort)
            if arguments['del']:
                treeIdKeys = parseGuessTreeIdentifiers(arguments['<xmlProteomeFile>'])           
                goStoreDel(treeIdKeys)
                exit(0)
            if arguments['add']:
                print(f"Connectin to goStore service to add following proteome(s){arguments['<xmlProteomeFile>']}")
                taxidTreeIter = loadUniversalTreesFromXML(\
                    arguments["<xmlProteomeFile>"],\
                    arguments["<owlFile>"])
                goStoreAdd(taxidTreeIter)
    
    elif arguments['pwas']:
        handshake(goApiHost, goApiPort)
        if arguments['server']:
            pwas_app = pwas_listen(goApiHost, goApiPort, arguments['vector'])
            pwas_app.run(debug=True, port=pwasApiPort)

        elif arguments['cli']:
            cli.run(aguments["expressed_protein_file"],\
                    arguments["delta_protein_file"],\
                    goApiPort, taxid,
                    asVector=arguments['vector'] # Check this test
            )


from pyproteinsExt import ontology
import uuid, os, pickle
import os.path
from .node import Node as createNode
from . import heap 

GO_ONTOLOGY = None

enumNS = {
            'biological process' : 'GO:0008150',
            'molecular function' : 'GO:0003674',
            'cellular component' : 'GO:0005575'
        }
        
def setOntology(owlFile=None, url=None):
    global GO_ONTOLOGY
    if not owlFile and not url:
        raise ValueError("file or url named parameters required")
    if owlFile:
        GO_ONTOLOGY = OntologyGO(owlFile=owlFile)
    else:
        GO_ONTOLOGY = OntologyGO(url=url)
    if GO_ONTOLOGY is None:
        raise ValueError(f"Could not set GO ontology from {owlFile} or {url}")

def ontologyDump(name, location):
    name = str( uuid.uuid4() ) if not name else name
    location = os.getcwd() if not location else location
    fName = location + '/' + name + '.sqlite3'
    print (f"Trying to open {fName}")
    ontology.default_world.set_backend(filename = fName)
    ontology.default_world.save()



# DAG implemnetaion requires memoizing visited node
# -> Any external accessor must be at the tree level, not at the node level avoid memoizing conflict
# Memoizing can be with a node heap or a node attribute
# We try the second one

# Check and raise Error at the 1st encountered cycle in
# the subtree of provided node 
import copy
def _checkCycle(rootNode, cNode, _path):
    path = copy.deepcopy(_path)
    path.append(cNode.ID)
    for child in cNode.children:
        if child.ID == rootNode.ID:
            pathString = '->'.join(path) + '->' + child.ID
            print (pathString)
            raise ValueError(f"{rootNode.name} has a cycle {pathString}")

def checkCycle(rootNode, verbose=False):
    if verbose:
        print(f"Checking {rootNode.ID}")

    for child in rootNode.children:
        _checkCycle(rootNode, child, [rootNode.ID])
    
    for child in rootNode.children:
        checkCycle(child)


class OntologyGO():
    def __init__(self, owlFile=None, url=None):
        if not owlFile is None:
            self.onto = ontology.Ontology(owlFile=owlFile)
        elif not url is None:
            self.onto = ontology.Ontology(url=url)
        else :
            raise("required parameter owlFile or url")
        
        print('Loaded')
        
    def getLineage(self, goID):
        lin = self.onto._getLineage(goID)
        if not lin:
            return None
        return [ (t.id[0], t.label[0]) for t in lin[:-1] ]


def ascend(cNode, nodeSet, rootSet):#;tree):
    """ Recursively look for the 1st ascendant node w/out parent
        
    """
    #print(f"-->{cNode.ID}")
    for _p in cNode.oNode.is_a:
        # non ascendant is_a element
        #http://www.ontobee.org/ontology/RO?iri=http://purl.obolibrary.org/obo/BFO_0000051
    
        if str(_p).startswith('obo.BFO_0000051'):
            continue
        
        try : # trying to dereference Restriction wrapper
            p = _p.value
        except AttributeError:
            p = _p

        # cNode is terminal node
        if str(p) == "owl.Thing" :
            if len(cNode.oNode.is_a)> 1:
                raise TypeError(f"Terminal node having multiple parent: {cNode}")
            #print(f"Adding {cNode} as root")
            rootSet.add(cNode)
            return

        #print(f"{p}")
        # the current pNode is already registred, we eventually register cNode as one of its children
        # But won't keep on ascending
        pNode =  createNode(p.id[0], p.label[0], p)
        toStop = True  if pNode in nodeSet else False
       
        # register current parent node
        pNode = nodeSet.add( pNode )
        pNode.isDAGelem = True
        #pNode.heap = tree.nodeHeap
        if not pNode.hasChild(cNode.ID):
            pNode.children.append(cNode)
        
        if not toStop:
            ascend(pNode, nodeSet, rootSet)#, tree)

def collapseTree(_root):
    root = copy.deepcopy(_root)
    ctHeap = heap.CoreHeap()
    root.children = [ n._collapseNode(ctHeap) for n in root.children ]
    return root
    
def insertLineage(root, lineage, eName):
    cNode = root
    for (goID, name) in reversed(lineage):
        mNode = cNode.hasChild(goID)
        if not mNode:
            #print(f"{goID},{name} NOT found under {cNode['ID']},{cNode['name']}")
            mNode = createNode(goID, name, None)
            cNode.children.append(mNode)
        else :
            #print(f"{goID},{name} found under {cNode['ID']},{cNode['name']}")
            pass
        cNode = mNode
        cNode.leafCount += 1
    cNode.eTag.append(eName)
    cNode.leafCount -= 1

def deserializeGoTree(fPickle, owlFile):
    global GO_ONTOLOGY
    if not GO_ONTOLOGY:
        setOntology(owlFile)
    
    fp = open(fPickle, 'rb')
    _self = pickle.load(fp)
    fp.close()
    for n in _self.walk():
        if n.oNode is None:
            print(f"Root found skipping::{n}\n")
            continue
        if isinstance(n.oNode, str):
            termID = n.oNode#.replace('obo.', '').replace('_', ':')
            termObj = GO_ONTOLOGY.onto.onto.search_one(id=termID)

            if termObj is None:
                raise TypeError(f"No GO Term matching::{termID}::From::{n}\n")
            n.oNode =  termObj

    return _self

def createGoTree(ns=None, proteinList=None, uniprotCollection=None, collapse=True) :
    if  ns is None:
        raise ValueError("Specify a namespace \"ns\"")
    if  proteinList is None:
        raise ValueError("Specify a list of proteins \"proteinList\"")
    if uniprotCollection is None:
        raise ValueError("Specify a collection of uniport elements \"uniprotCollection\"")

    print(f"Extracting {ns} ontology")

    xpGoTree = AnnotationTree(ns, collapse=True)
    xpGoTree.extract(proteinList, uniprotCollection)
    return xpGoTree

def load(baseData):
    """Deserialize tree as dict structure fetched from api
        parameter is a shallow dict structure wich is assumed to 
        feature a single value with an empty "is_a" array ie a single root element
    {
        "0000":{"children":["GO:0008150"],"eTag":[],"is_a":[],"name":"root"},
        "GO:0008150": {"children":["GO:0051179","GO:0009987","GO:0008152","GO:0050896"],"eTag":[],"is_a":["0000"],"name":"biological_process"},
        ...
    }
    """
    
    flatNodeData = spawn(baseData)
    root = wire(flatNodeData, baseData)
    maybeNS = root.children[0].name.replace('_', ' ')
    
    t = AnnotationTree(maybeNS, collapse=True)
    t.root = root
    
    t.isDAG = True
    t.collapsable = True # Arbitrary
    t.leafCountUpdate()
    
    return t
        
def spawn(strData):
    """Utility deserialize function
        Generate a shallow dictionary of Node objects
    """
    nodeData = {}
    for nKey, nDatum in strData.items():
        nodeData[nKey] = createNode(nKey, nDatum['name'])
        nodeData[nKey].eTag = nDatum['eTag']
        nodeData[nKey].isDAGelem = nDatum['isDAGelem']
    #print(f"spawn Nodedata has {len(list(nodeData.keys()))} elem")
    
    return nodeData

def wire(nodeData, strData, mayDropOccur=False):
    """Utility deserialize function
        fill Node reference across provided dictionary on Node objects
        using Node identifiers stored in 2nd paramter
        return the root of corresponding tree
    """
    root=None
    for nKey, nObj in nodeData.items():
        nRefStr = strData[nKey]
        if len(nRefStr['is_a']) == 0:
            if not root is None:
                raise KeyError("Multiple roots")
            root = nObj
        for childID in nRefStr['children']: # Children missing reference are allowed in the context of drop based tree reconstruction
            if not childID in nodeData:
                if mayDropOccur:
                    continue
                else:
                    raise KeyError(f"while editing children reference, {childID} not found in shallow node dictionary")
            nObj.children.append(nodeData[childID])

        for parentID in nRefStr['is_a']: # parent missing reference are forbidden for now
            nObj.is_a.append(nodeData[parentID])
    
    return root

   



class AnnotationTree():
    def __init__(self, annotType,  collapse=False):
        
        if annotType not in enumNS:
            raise KeyError (f"annotation type \"{annotType}\" is not allowed ({enumNS}) {{ {enumNS.keys()} }}")

        self.collapsable = collapse
        self.isDAG = False
        #self.nodeHeap = heap.CoreHeap()
        self.root = createNode('0000', 'root') 
        #self.root.heap = self.nodeHeap
        self.NS = (annotType, enumNS[annotType])

    def vectorize(self):
        """Create a dictionary representation of passed tree content, getting rid of tree topology 
        in exchanged for a vectorizable data structure of node terms lited under the "elements" list.
        All dictionary key,value are strings or integer for easy serialization.
        The returned dictionary layout is the following:

        { 
            "registry" : [uniprotID,],
            "terms" : { "GOid" : [ elements.index, ... ], }
        } 

        Parameters
        ----------
        aTree: a tree.AnnotationTree object.

        Returns
        -------
        A dictionary of tree nodes
        """
        vHeap = heap.VHeap()
        for n in self.walk():
            vHeap.add(n, None)
        return vHeap.asDict

    # Serializing without any ONTOLOGY references
    def f_serialize(self):
        fHeap = heap.FHeap()
        cnt = 0
        for cNode in self.traverse():
            fHeap.add(cNode, None)
            for node in cNode.children:
                fHeap.add(node, cNode)
            cnt += 1
        print(f"{cnt} nodes traversal")
        print(f"heap dimensions:: nodes:{fHeap.dimensions[0]} childrenRefCount :{fHeap.dimensions[1]}")

        return fHeap
    
    # Serialization requires pickling of object instance
    # Ontology related term will be saved as string
    # The ontology object will have to be reimported for deserialization
    def makePickable(self):
        _self = copy.deepcopy(self)
        for n in _self.walk():
            n.oNode = str(n.oNode).replace('obo.', '').replace('_', ':') if not n.oNode is None else None
        return _self
    
    def extract(self, *args):
        self.read_DAG(*args)

    def read_DAG(self,uniprotIDList, uniprotCollection):      
        """ Cross GO Ontology with supplied uniprotID List
            to create the minimal GO DAG containg all GO terms featured by uniprot collection
        """
        self.isDAG = True
        global GO_ONTOLOGY
        if GO_ONTOLOGY is None:
            print("Please set GO_ONTOLOGY")
            return
      
        
        ontologyNode = GO_ONTOLOGY.onto.onto.search_one(id=self.NS[1])
        if not ontologyNode:
            raise ValueError(f"id {enumNS[annotType]} not found in ontology")
      #  self.root.children.append( Node(enumNS[annotType], annotType, oNode=ontologyNode) )

        nodeSet = heap.CoreHeap()
        rootSet = heap.CoreHeap()
        def setSentinelChar():
            """ Returns the letter used by GO to prefix its term depending on namespace"""
            if self.NS[0] == 'biological process':
                return 'P'
            elif self.NS[0] == 'molecular function':
                return 'F'
            return 'C'
        
        goNSasChar = setSentinelChar()
        disc = 0
        for p in uniprotIDList:
           
            u = uniprotCollection.get(p)
            bp = []
            for goTerm in u.GO:
                if goTerm.term.startswith(f"{goNSasChar}:"):
                    bp.append(goTerm.id)
            if not bp:
                disc += 1
                #print(f"Added {p} provided not GO annnotation (current NS is {self.NS[0]})")

            for term in bp:
                cLeaf = GO_ONTOLOGY.onto.onto.search_one(id=term)
                if not cLeaf:
                    cLeaf = GO_ONTOLOGY.onto.onto.search_one(hasAlternativeId=term)
                    if not cLeaf:
                        #print(f"{u.GO[annotType][term]} {term} not found")
                        raise KeyError(f"{u.GO[self.NS[0]][term]} {term} not found")
                # Add a new node to set of fetch existing one
                bottomNode = nodeSet.add( createNode(cLeaf.id[0], cLeaf.label[0], cLeaf) )
                bottomNode.eTag.append(p)
                bottomNode.isDAGelem = True
                #bottomNode.heap = self.nodeHeap
                #print(f"rolling up for {bottomNode.ID}")
                ascend(bottomNode, nodeSet, rootSet)#, self)
                
                
        #if len(rootSet) > 1:
        #    raise ValueError(f"Too many roots ({len(rootSet)}) {list(rootSet)}")
        for n in rootSet:
            if n.ID == self.NS[1]:
                self.root.children.append(n)
        if self.collapsable:
            print("Applying true path collapsing")
            self.root = collapseTree(self.root)
            #self.nodeHeap = self.root.heap
            
            
        n, ln, l, p = self.dimensions
        print(f"{n} GO terms, {ln} children_links, {l} leaves, {p} proteins ({disc} discarded)")

    def read(self,uniprotIDList, uniprotCollection):      
        global GO_ONTOLOGY
        if GO_ONTOLOGY is None:
            print("Please set GO_ONTOLOGY")
            return
        
        self.root.children.append( createNode(self.NS[1], self.NS[0]) )
               
        i=0
        for p in uniprotIDList:
            u = uniprotCollection.get(p)
            bp = list((u.GO[self.NS[0]]).keys())
            if bp:
                i += 1          
            for term in bp:
                lineage = [(term, u.GO[self.NS[0]][term])] + GO_ONTOLOGY.getLineage(term)
                insertLineage(self.root, lineage, p)
                
        print(f"Annotation {self.NS[0]} extracted from {i} / {len(uniprotCollection)} uniprot entries parsed")
        
        if self.collapsable:
            print("Applying true path collapsing")
            self.root = collapseTree(self.root)
          #  self.nodeHeap = self.root.heap
    
    @property 
    def dimensions(self):
        nNodes = 0
        nLinks = 0
        for n in self.root.walk():
            nNodes += 1
            nLinks += len(n.children)
        leafTotal = self.root.getMembers()
        leafTotal_nr = set(leafTotal)

        return nNodes, nLinks, len(leafTotal), len(leafTotal_nr)

    @property
    def proteins(self):
        return set(self.root.getMembers())

    def traverse(self):
        return self.root.traverse() 
    
    def walk(self, **kwargs):       
        return self.root.walk(**kwargs) 
    
    def as_json(self): # Should be collapsible
        return json.dumps(self.root)

    def as_newick(self, collapse=True):
        return '(' + self.root._as_newick() + ');'

    def getByName(self, name):
        n = self.root.getByName(name)
        if not n:
            raise KeyError(f"No node named \"{name}\" in current tree")
        return n

    def getByID(self, ID):
        n = self.root.getByID(ID)
        if not n:
            raise KeyError(f"No node w/ ID \"{ID}\" in current tree")
        return n

    def collapse(self):
        t = AnnotationTree(self.NS[0])
        t.isDAG = self.isDAG
      

        t.root = collapseTree(self.root)
       # t.nodeHeap = t.root.heap
        t.collapsable = True
        return t

    def getMembersByName(self, name, nr=True):
        node = self.root.getByName(name)
        m = node.getMembers()
        if nr:
            s = set(m)
            return list(set(m)) 

        return list(m)
       
    def getMembersByID(self, name, nr=True):
        node = self.root.getByID(name)
        #print(node)
        m = node.getMembers()
        if nr:
            return list(set(m))
        
        return list(m)

    def getMembers(self):
        return self.root.getMembers(nr=True)

    def newRoot(self,  name=None, ID=None):
        if not name and not ID:
            raise ValueError("Provide name or ID")

        t = AnnotationTree(self.NS[0])
        t.isDAG = self.isDAG
        t.collapsable = self.collapsable

        field = "name" if not name is None else "ID"
        value = name if not name is None else ID
        fn = self.getByName if not name is None else self.getByID

        nr = fn(value)

        if not nr:
            raise KeyError(f"No node with {field} {value} in mother tree")
        t.root.children.append(nr)
        
        return t

# A node may effectively be tested more than once
# Given the operation cost associated with False predicate (terminate&return None), heap is unecessary
# Given the operation cost associated with True predicate we effectively test the below tree once again
    def _drop(self, predicate, noCollapse=False):
        t = AnnotationTree(self.NS[0])
        t.isDAG = self.isDAG
        t.collapsable = self.collapsable
 
        noDropHeap = heap.CoreHeap() # To store success and avoid restest subtree

        t.root = copy.deepcopy(self.root)
        t.root.children = [ c for c in t.root.children if c._mayDrop(predicate, noDropHeap) ]

        if not noCollapse:
            t = t.collapse()
        t.leafCountUpdate()

        return t
    def drop(self, predicate, noCollapse=False, noLeafCountUpdate=False):
        """Returns a subtree with all nodes matching parameter predicate function
        """
        noDropHeap = heap.FHeap() # To store success and avoid restest subtree
        noDropHeap.add(self.root, None)
        # start recursive
        for _ in self.root.children:
            _.mayDrop(predicate, self.root, noDropHeap)
        
        # Build results tree from the flat dictionary of surviving nodes
        baseData = noDropHeap.asDict
        flatNodeData = spawn(baseData)
        root = wire(flatNodeData, baseData, mayDropOccur=True)
        
        t = AnnotationTree(self.NS[0])
        t.isDAG = self.isDAG
        t.collapsable = self.collapsable
        t.root = root
        if not noCollapse:
            t = t.collapse()
        if not noLeafCountUpdate:
            t.leafCountUpdate()

        return t

    def leafCountUpdate(self):
        lcHeap = heap.CoreHeap()

     #   if self.isDAG:
     #       print(f"clearing heap")
     #       self.nodeHeap.clear()
     #       print(f"heap size: {len(self.nodeHeap)}")
        for c in self.root.children:
            c._leafCountUpdate(lcHeap)

    def as_DAG_strat(self):
        return self.root.as_DAG_strat()



from flask import Flask, jsonify, abort, request, make_response, Response
from ..data_objects import CulledGoParametersSchema as goParameterValidator
from ..data_objects import loadUnivGO
from .cache import setCacheType, delTreeByTaxids, storeTreeByTaxid, getTaxidKeys, getUniversalTree
from .cache import getCulledVector, storeCulledVector, getUniversalVector
from .cache import clear as clearStore
from .cache import buildUniversalVector, listTrees, listVectors, listMissUniversalVector, listCulled
from .cache import status as storeStatus
from decorator import decorator
# IF OFFLOADABLE BLOCKING ROUTES KEEP ON SHOWING UP
# https://blog.miguelgrinberg.com/post/using-celery-with-flask

import time

from multiprocessing import Process, Value, Semaphore

bSemaphore = None
C_TYPE = None
_MAIN_ = False

def bootstrap(newElem=None, cacheType='local',\
    clear=False, _main_=False, **kwargs):
    global _MAIN_, C_TYPE
    
    if _main_:
        print("Bootstraping main process")
        _MAIN_ = _main_
        global bSemaphore
        bSemaphore = Semaphore(1) # bleeding eyes
        setCacheType(cacheType, **kwargs)
        C_TYPE = cacheType
        if clear:
            clearStore()

        if newElem:
            print("Boostraping unigo store with following new ressources")
            for taxid, tree in newElem:
                print(f"\tAdding{taxid}:{tree}")
                storeTreeByTaxid(tree, taxid)

        tcount, vcount, _vcount = storeStatus()
        print(f"Database content:\n\t{tcount} trees, {vcount} vectors, {_vcount} culled")

        app = Flask(__name__)

        app.add_url_rule('/ping', 'ping', ping)
        
        app.add_url_rule('/taxids', 'view_taxids', view_taxids)
        
        app.add_url_rule('/unigo/<taxid>', 'view_unigo', view_unigo, methods=['GET'])
        
        app.add_url_rule('/vector/<taxid>', 'view_vector', view_vector, methods=['GET'])
        
        app.add_url_rule("/vector/<taxid>", 'view_culled_vector', view_culled_vector, methods=['POST'])
        
        app.add_url_rule('/add/unigo/<taxid>', 'add_unigo', add_unigo, methods=['POST'])

        app.add_url_rule('/del/unigo/<taxid>', 'del_unigo', del_unigo, methods=['DELETE'])

        app.add_url_rule('/build/vectors', 'build_vectors', build_vectors)

        app.add_url_rule('/list/<elemType>', 'list_elements', list_elements)

        return app

def list_elements(elemType):
    if elemType == 'vectors':
        return jsonify({ 'vectors': listVectors() })
      
    elif elemType == 'trees':
        return jsonify({ 'trees': listTrees() })
    
    elif elemType == 'culled':
        return jsonify({ 'culled': listCulled() })


    print(f"Unknwon element type {elemType} to list")
    abort(404)

# CPU intensive route -> offloading it to multiprocess
# using semaphore to prevent concurrency popup on multiple requests
def build_vectors():
    if _MAIN_:
        missUniversalVector = listMissUniversalVector()
        if not missUniversalVector:
            return jsonify({"status": "nothing to build"}), 200

        # global bSemaphore exists
        if bSemaphore.acquire(block=False):
            print(f"bSemaphore is acquired")
            p = Process(
                target=_buildUniversalVector,
                args=(bSemaphore, C_TYPE),
                daemon=True
            )
            p.start()
            return jsonify({
                "status": "starting",
                "targets" : missUniversalVector
            }), 200
        
        print(f"bSemaphore is locked")   
        return jsonify({"status": "running"}), 202
    else: 
        print("##### Twilight zone ######") 


@decorator
def semHolder(fn, _bSemaphore, cacheType, *args, **kwargs):
    #print("Decrorator start")
    setCacheType(cacheType)
    #time.sleep(10)
    fn(*args, **kwargs)
    _bSemaphore.release()
    print(f"Relasing bSemaphore")

@semHolder
def _buildUniversalVector(*args,**kwargs):
    buildUniversalVector()

def ping():
    return "pong"

def view_taxids():
    return str( getTaxidKeys() )

def view_unigo(taxid):
    try:
        tree = getUniversalTree(taxid)
        return jsonify(tree.serialize())
    except KeyError as e:
        print(e)
        abort(404)

##################################################
# ACCESSORS to taxid specific goterms as vectors
##################################################
# GET route, return universal vector
def view_vector(taxid):
    try:
        taxidVector = getUniversalVector(taxid)
    except KeyError as e:
        print(e)
        abort(404)
    
    return jsonify(taxidVector)
# POST route
# Presuming live culling is not CPU intensive enough to
# promote multiprocess offloading:: CHECK IT
def view_culled_vector(taxid): 
    try:
        taxidVector = getUniversalVector(taxid)
    except KeyError as e:
        print(e)
        abort(404)
    try:
        data = request.get_json()
        _goParameterValidator = goParameterValidator()
        goParameter = _goParameterValidator.load(request)
        cmin, cmax, fmax = ( goParameter['minCount'],\
            goParameter['maxCount'], goParameter['maxFreq'])
    except Exception as e:
        print(e)
        print(f"Malformed GO parameters\n=>{data}")
        abort(400)

    try:
        _ = getCulledVector(taxid, cmin, cmax, fmax)
        print(f"Culled vector found ! [{taxid} {cmin} {cmax} {fmax}]")
    except KeyError : # Culled vector not in cache, build&store
        print(f"Building culled vector... [{taxid} {cmin} {cmax} {fmax}]")
        _ = { # Do we got it all ? Dump it once
            'registry' : taxidVector['registry'],
            'terms' : {  goID : goTerm for goID, goTerm in taxidVector['terms'].items()\
                                                    if len(goTerm['elements']) >= cmin and\
                                                       len(goTerm['elements']) <= cmax and\
                                                       goTerm['freq']          <= fmax \
            }
        }
        storeCulledVector(_, taxid, cmin, cmax, fmax)

    return jsonify(_)

def add_unigo(taxid):
    """ Add a unigo object through client API
        Parameters:
        ---------- 
        taxid: ncbi taxid 
        json payload: ..Univgo.serialize() dict
    """
    try:
        univGoTree = loadUnivGO(request.get_json())
        storeTreeByTaxid(univGoTree, taxid)
        return jsonify({"taxid" : "insertion OK"})
    except KeyError as e:
        print(f"{taxid} already exist in database or malformed request, reject insertion\n=>{e}")
        abort(403)
    except Exception as e:
        print(f"add unigo internal error:{e}")
        abort(500)

def del_unigo(taxid):
    try:
        delTreeByTaxids([taxid])
    except KeyError:
        print(f"{taxid} not found in database, nothing to delete")       
        abort(404) 
  
    return jsonify({"taxid" : "deletion OK"})


from .utils import loadUniprotCollection
from . import Unigo as createGOTreeTest

DEFAULT_PROTEOME=f"{os.path.dirname(os.path.abspath(__file__))}/data/uniprot-proteome_UP000000807.xml.gz"
def run(nDummy=50, nTop=10, target='fisher'):
    print(f"Testing local implementation of {target}")
    
    print(f"Loading test proteome {DEFAULT_PROTEOME}")
    uTaxid, uColl  = loadUniprotCollection(DEFAULT_PROTEOME, strict=False)
    deltaUniprotID = generateDummySet(uColl, nDummy)
    
    print("Creatin unigo Tree")        
    unigoTree = createGOTreeTest(backgroundUniColl = uColl,
                                proteinList       = expUniprotID,
                                fetchLatest       = False)
    
    if target == 'fisher' or target == 'convert':
        print("Computing ORA")
        rankingsORA = unigoTree.computeORA(deltaUniprotID)
        print(f"Test Top - {nTop}\n{rankingsORA[:nTop]}")

    if arguments['convert']:
        print("Testing tree serialization")
        d = unigoTree.tree.f_serialize()
        print(d.asDict)



from .utils import loadUniprotIDsFromCliFiles, unigo_tree_from_api
import json
from .stat_utils import applyOraToVector
from . import uloads as createGOTreeFromAPI

# NO CULLING RESSOURCE USAGE, TO IMPLEMENT
def run(expUniprotIdFile, deltaUniprotIdFile, goApiHost, goApiPort, taxid, method="fisher", asVector=True):
    expUniprotID, deltaUniprotID = loadUniprotIDsFromCliFiles(\
                                            expUniprotIdFile,\
                                            deltaUniprotIdFile
                                            )
    if asVector: # Vector ora     
        print("Running the vectorized ora")
        resp = utils.unigo_vector_from_api(goApiHost, goApiPort, taxid)
        if resp.status_code != 200:
            print(f"request returned {resp.status_code}")
        
        vectorizedProteomeTree = json.loads(resp.text)
        res = applyOraToVector(vectorizedProteomeTree, expUniprotID, deltaUniprotID, 0.5)
        print(res)

    else:# Tree ora       
        resp = utils.unigo_tree_from_api(goApiHost, goApiPort, taxid)
        if resp.status_code != 200:
            print(f"request returned {resp.status_code}")  

        unigoTreeFromAPI = createGOTreeFromAPI(resp.text, expUniprotID)
        x,y = unigoTreeFromAPI.dimensions
        assert not unigoTreeFromAPI.isExpEmpty
        if method == "fisher":
            print("Computing ORA")
            rankingsORA = unigoTreeFromAPI.computeORA(deltaUniprotID)
            print(f"Test Top - {nTop}\n{rankingsORA[:nTop]}")
    


from .tree import setOntology, createGoTree, load
from . import stat_utils
import os
import json

import datetime

print("DVL PYPROTEINS_EXT PKG", datetime.datetime.now())

"""
Minimal ontology tree of a uniprot collection

Merging Uniprot Object Collection 
Consrtuctor

"""

def prune(unigoObj, predicate):
    """Create a Unigo object by Pruning the supplied Unigo instance

        Parameters
        ----------
        unigoObj : 
        predicate: predicate function applied to prodived Unigo to drop or keep its elements
    """
    _tree          = unigoObj.tree.drop(predicate)
    _tree_universe = unigoObj.tree_universe.drop(predicate)
    return Unigo( previous=(_tree, _tree_universe) )


def chop(unigoObj, name=None, ID=None):
    """Create a Unigo object with its root extracted from the provided Unigo instance

        Parameters
        ----------
        unigoObj : 
        name: name of the node the look for in provided unigoObj. It will be the root of the returned Unigo
        ID: ID of the node the look for in provided unigoObj. It will be the root of the returned Unigo
    """
    
    _tree          = unigoObj.tree.newRoot(name=name, ID=name)
    _tree_universe = unigoObj.tree_universe.newRoot(name=name, ID=name)
    return Unigo( previous=(_tree, _tree_universe) )

def vloads(data):
    """Deserialize a Univgo object (w/out uniprotColl reference)
    """
    _data = json.loads(data)
    return  Univgo(serial=_data)
 
    # Update leaf count
class Univgo:
    def __init__(self, serial = None, 
                       uniColl = None, 
                       owlFile=None, fetchLatest=True,
                       ns="biological process"):#, **kwargs):
        """Create a Univgo object, made to manipulate whole proteome ontology

        Parameters
        ----------
        owlFile : path to the ontology owl file.
        uniColl : collection of uniprot objects, defining the background population
        [Options]
        ns : a subset of ontology, default:biological process

        """
        if serial:
            self.single_tree     = load(serial["single_tree"])
            self.omega_uniprotID = serial["omega_uniprotID"]
            self.ns              = serial["ns"]
            return
        
        if uniColl is None :
            raise ValueError(f"parameters uniColl required")
        try :
            if owlFile is None and fetchLatest:
                print("Fetching ontology")
                setOntology(url="http://purl.obolibrary.org/obo/go.owl")#, **kwargs)
            elif not owlFile is None:
                setOntology(owlFile=owlFile)
            else:
                print("Using package release ontology")
                setOntology(owlFile=f"{os.path.dirname(os.path.abspath(__file__))}/data/go.owl")
        except Exception as e:
            print(f"Could not create ontology")
            print(e)
            raise TypeError("Failed creating Go tree")
        self.omega_uniprotID = [ k for k in uniColl.keys() ]
        self.single_tree = createGoTree(         ns = ns,
                                  proteinList       = self.omega_uniprotID, 
                                  uniprotCollection = uniColl)
        self.ns = ns

    def serialize(self):
        return {
            "single_tree"     : self.single_tree.f_serialize().asDict,
            "omega_uniprotID" : self.omega_uniprotID,
            "ns"              : self.ns    
        }

    def vectorize(self):
        data = self.single_tree.vectorize()
        print(len(data["registry"]))
        for uniprotID in set(self.omega_uniprotID) - set(data["registry"]):
            data["registry"].append(uniprotID)
        print(len(data["registry"]))
        return data

    @property
    def dimensions(self):
        return self.single_tree.dimensions

def uloads(data, proteinIDList):
    """Generate a Unigo tree Object
    """
    _data=json.loads(data)

    tree_universe = Univgo(serial=_data).single_tree
    # Set predicate function to retain node in universal tree whih are path to provided proteins
    def keep_proteinID(node):
        return set(node.getMembers()) & set(proteinIDList)
    
    # Obtaine the tree
    exp_tree = tree_universe.drop(keep_proteinID, noCollapse=True, noLeafCountUpdate=True)
    # Reset eTags to keep only provided proteinIDList
    for n in exp_tree.walk():
        n.eTag = list(  set(proteinIDList) & set(n.eTag) )
    # update tree topology and leaf counts
    tree = exp_tree.collapse()
    exp_tree.leafCountUpdate()

    return Unigo(previous=(tree, tree_universe))

class Unigo:
    def __init__(self, previous = None, 
                       backgroundUniColl = None, 
                       proteinList       = None, 
                       owlFile=None, fetchLatest=True,
                       ns="biological process"):#, **kwargs):
        """Create a Unigo object

        Parameters
        ----------
        owlFile : path to the ontology owl file.
        backgroundUniColl : collection of uniprot objects, defining the background population
        proteinList : A list of uniprot identifiers
        [Options]
        ns : a subset of ontology, default:biological process

        TODO: Check if all protein list are members of bkgUniCol

        """
        if previous:
            self.tree, self.tree_universe = previous
            return
        
        if backgroundUniColl is None or  proteinList is None:
            raise ValueError(f"parameters backgroundUniColl and proteinList are required")
        try :
            if owlFile is None and fetchLatest:
                print("Fetching ontology")
                setOntology(url="http://purl.obolibrary.org/obo/go.owl")#, **kwargs)
            else:
                print("Using package release ontology")
                setOntology(owlFile=f"{os.path.dirname(os.path.abspath(__file__))}/data/go.owl")
        except Exception as e:
            print(f"Could not create ontology")
            print(e)
        
        self.tree = createGoTree(          ns       = ns,
                                  proteinList       = proteinList, 
                                  uniprotCollection = backgroundUniColl)

        proteinUniverseList = [ k for k in backgroundUniColl.keys() ]
        self.tree_universe = createGoTree(                ns = ns, 
                                           proteinList       = proteinUniverseList, 
                                           uniprotCollection = backgroundUniColl)

    @property
    def isExpEmty(self):
        return self.tree.dimensions[0] == 1
    
    @property
    def dimensions(self):
        return (self.tree.dimensions, self.tree_universe.dimensions)
    
    def walk(self, type=None):
        """Iterate over the tree, by default the experimental one"""
        if type is None:
            for node in self.tree.walk():
                yield node

    def computeORA(self, proteinIDList, rootPathway=None, ns="biological process", verbose=False):
        """Recursively computes the Fisher exact test of all pathways 
            based on corrresponding intersections between the following two categories
                    | Pa  | non_PA |
             -----------------------
                SA  |     |        |
              nonSA |     |        |
            Where,
                PA is the set of protein annotated in a pathway, extracted from the whole proteome tree
                SA is the set of proteins with altered experimental quantities: the <proteinIDList> parameter

        """
        #_ns = ns.replace(' ', '_')
        pathWayXP = self.tree.getByName(ns)
        if rootPathway:
            pathWayXP = self.tree.getByName(rootPathway)
            if pathWayXP is None:
                print(f"No GO term named{None} was found in current ontology (maybe not populated)")
                return None
        
        pathWayBKG = self.tree_universe.getByName(ns)

        oraFisher, oraCDF = stat_utils.computeORA_BKG( pathWayXP,      # Pathway Node with all experimental proteins
                                                       proteinIDList,  # Set of protein w/ altered xptl qtty: observation counts
                                                       pathWayBKG,     # Available sets of protein in the pathways: hidden success counts [FROM PROTEOME]
                                                       verbose=verbose)
        return ORA_report(self.tree, oraFisher, oraCDF)
# Define rich view of stat results for notebook ?
class ORA_report:
    def __init__(self, xpTree, scoreF, scoreC):
        
        self.scoreF =  [ ORA_score(_) for _ in sorted(scoreF, key=lambda x:x[0]) ]
        self.scoreC =  [ ORA_score(_) for _ in sorted(scoreC, key=lambda x:x[0]) ]

    def __getitem__(self, x):
        return self.scoreF[x]

    @property
    def json(self, topN=0):
        return {"scoreF" : [{"go_term" : _.cNode.name, "p_val": _.score, "proteins": _.proteins } for _ in self.scoreF],\
            "scoreC":[{"go_term" : _.cNode.name, "p_val": _.score, "proteins": _.proteins } for _ in self.scoreC]}

class ORA_score:
    def __init__(self, scoreElem):
        self.score, self.cNode = scoreElem
        self.proteins = self.cNode.getMembers()
    def __repr__(self):
        return f"{self.cNode.name} {self.score} {self.proteins}"

#import json

#res = {}

#for node in goTreeObjExp.walk():
#    if node.pvalue:
#         res[ node.name ] = {
#             "name"          : node.name,
#             "pvalue"        : node.pvalue,
#             "proteineTotal" : node.getMembers(nr=True),
#             "proteineSA"    : list ( set (saList) & set(node.getMembers(nr=True)) )
#         }
        
#with open('TP_ORA.json', 'w') as fp:
#    json.dump(res, fp)
#res